function gill_cell_ga_search_1
clc;clear;close all;

%Initialize search targets.
filename = 'test_data_search';
low_lim = [0 -2.5 -1 -1 3];
high_lim = [3 2 3 0.5 4];
target = 'gill_target_lq';
NUM_PROC=4;

TARGET = load(sprintf('search_target_data/%s.mat',target));

%Set up globals tracking parameter values throughout search.
global search_step_params
global search_step_fvals
global search_step_target
global search_step_mode
search_step_params = {};
search_step_fvals = {};
search_step_target = [];
search_step_mode = {};



target_hist = TARGET.histout;
target_mean = TARGET.meanout;
target_bins = TARGET.binout;

p_sim_stoch = struct;
p_sim_stoch.max_bin = size(target_hist,3)-1;
p_sim_stoch.elongationmode = 'constant';

timevec = TARGET.timevec;
p_phys_stoch = TARGET.p_phys_stoch;

onoff = 'on';
units = 'min';

%Set up search parameters.
num_reduc = 5;
num_vars = length(low_lim);
population_size = 200;

opt_time_bounds = [timevec(1) 3.5;...
    timevec(1) timevec(end);...
    timevec(1) timevec(end);...
    timevec(1) timevec(end);...
    timevec(1) 3.5];
opt_metric = {'cdf','mean','cdf','mean','cdf'};
opt_target = [1 3 3 1 1];

radius = [0.14 0.12 0.11 0.12];
multiplier = 3;
 
%Set up simulation coarseness for each stage of search
Nuc = ceil(5301/15)*[1 1 1 1 1];
% Cells = [1500 500 700 900 3500];
Cells = [20 50 50 50 50];

%Set up GA options
num_generations = [ 5 8 8 8 5];
crossover_ratio = [0.5 0.25 0.23 0.16 0.13];

ga_options_init = gaoptimset(...
    'Display','none',...
    'Vectorized','on',...
    'PopulationSize',population_size,...
    'FitnessScalingFcn',@fitscalingprop,...
    'EliteCount',0,...
    'CrossoverFraction',1,...
    'SelectionFcn',@selectionstochunif);

%Create variables to track runtime and results
run_time = NaN(num_reduc,1);

output_cell = cell(num_reduc,1);
fval_cell = cell(num_reduc,1);

scaling_mode_preset = {'uniform','uniform','uniform','uniform','uniform'};
cell_scaling = [10 10 10 10 1];
pop_scaling = [10 5 5 5 1];

for i = 1:num_reduc

    time_bounds = opt_time_bounds(i,:);
    
    %Set up weight function for scalar objective function.
    if strcmp(opt_metric{i},'mean')
        weight_func = ones(size(timevec));
    elseif strcmp(opt_metric{i},'cdf') && opt_target(i)==1
        weight_func = exp(-timevec/timevec(end)*10);
    elseif strcmp(opt_metric{i},'cdf') && opt_target(i)==3
        weight_func = exp((timevec/timevec(end)-1)*3);
    else 
        warning('Time weight function not set, using default uniform weighing!');
        weight_func = ones(size(timevec));
    end
    
    if strcmp(opt_metric{i},'mean')
        likeness_func = @likeness_calc_lsq;
    elseif strcmp(opt_metric{i},'cdf')
        likeness_func = @likeness_calc_emd;
    end
        
    

    CELL_SCALING = cell_scaling(i);
    POP_SCALING = pop_scaling(i);
    scaling_mode = scaling_mode_preset{i};
    cell_vec = NaN(num_generations(i),1);
    
    pop_vec = NaN(num_generations(i),1);
     
    
    if strcmp(scaling_mode,'linear')
        %For linear increase in # cells / decrease in # param sets
        cell_vec(1:end) = ceil(linspace(Cells(i)/CELL_SCALING,Cells(i),num_generations(i)));
        pop_vec(1:end) = ceil(linspace(population_size*POP_SCALING,population_size,num_generations(i)));
    elseif strcmp(scaling_mode,'init_coarse')
        %For nonlinear increase: sample a lot of param sets at first
        scaling_factor_fine = 4;
        
        cell_vec(1) = ceil(Cells(i)/CELL_SCALING);
        cell_vec(2:end) = ceil(linspace(cell_vec(1)*scaling_factor_fine,Cells(i),num_generations(i)-1));
        
        pop_vec(1) = population_size*POP_SCALING;
        pop_vec(end) = population_size;
        pop_vec(2:end) = ceil(linspace(pop_vec(1)/scaling_factor_fine,population_size,num_generations(i)-1));
    elseif strcmp(scaling_mode,'uniform')
        %Constant population size
        cell_vec(1:end) = Cells(i);
        pop_vec(1:end) = population_size;
    end
    
    ga_options = gaoptimset(...
        ga_options_init,...
        'Generations',1,...
        'CrossoverFcn',{@crossoverheuristic,(1+crossover_ratio(i))});
    
    timer_obj = tic;
    for gen = 1:num_generations(i)
        
        if gen>1 || i>1
%             start_points
            %Visualize the population
            n_start_pts = size(start_points,1);
%             figure(2);clf;
%             scatter3(start_points(:,1),start_points(:,2),start_points(:,3),10,'k','filled');hold on;
            while size(start_points,1) < pop_vec(gen)
                %If there are too few points for next step, use linear
                %interpolation
                wt_lin_comb = rand;
                appendix = wt_lin_comb*start_points(randi(n_start_pts,1),:) + ...
                    (1-wt_lin_comb)*start_points(randi(n_start_pts,1),:);
                start_points = [start_points;...
                   appendix];
            end
            %Compare interpolated points with initial points
%             scatter3(start_points(:,1),start_points(:,2),start_points(:,3),15,'r');
        
            ga_options = gaoptimset(...
                ga_options,...
                'InitialPopulation',start_points);
            if gen > 1
                ga_options = gaoptimset(...
                ga_options,...
                'InitialScores',initial_scores);
                
            end
            
            
        end
        ga_options = gaoptimset(...
            ga_options,...
            'PopulationSize',pop_vec(gen));

        
        if i > 1
            fun_search = @(kvar_log) fun_search_variable(...
                kvar_log,timevec,p_phys_stoch,p_sim_stoch,...
                units,onoff,time_bounds,...
                target_hist,target_mean,target_bins,...
                likeness_func,...
                Nuc(i),cell_vec(gen),NUM_PROC,...
                weight_func,opt_target(i),opt_metric{i},low_lim,high_lim,discount_func);
                %Use cost increase function if after first stage
        else
            fun_search = @(kvar_log) fun_search_variable(...
                kvar_log,timevec,p_phys_stoch,p_sim_stoch,...
                units,onoff,time_bounds,...
                target_hist,target_mean,target_bins,...
                likeness_func,...
                Nuc(i),cell_vec(gen),NUM_PROC,...
                weight_func,opt_target(i),opt_metric{i},low_lim,high_lim);            
        end

        %Track number of trajectories
        fprintf('stage %i, gen %i, %i cells, %i parameter sets, %i trajectories\n',...
            i,gen,cell_vec(gen),ga_options.PopulationSize,cell_vec(gen)*ga_options.PopulationSize);
        

%         tic
        [output_point,fval_point,~,~,final_pop,final_scores] = ga(...
            fun_search,num_vars,[],[],[],[],low_lim,high_lim,[],ga_options);
%         toc
            
        
        if gen < num_generations(i)
            frac_points_retained = pop_vec(gen+1)/pop_vec(gen);
            
            if frac_points_retained<1
                Q = quantile(final_scores,frac_points_retained);
%                 Q
                point_filter = final_scores<Q;
%                 point_filter
            %If this part gives an error, the metric is giving degenerate results.
            %Perhaps try more cells or a different metric.
            %Or retain all points = use uniform setting for population.
                final_pop = final_pop(point_filter,:);
                final_scores = final_scores(point_filter);
            end
            
            
        end
        start_points = final_pop;
        initial_scores = final_scores;
    end

    run_time(i) = toc(timer_obj);
    
    if i < num_reduc
        
        good_fit_filter = true(size(final_scores));
        good_fit_pts = final_pop(good_fit_filter,:);
        
        discount_func = @(test_pts) err_compensator(test_pts,...
            good_fit_pts,radius(i),multiplier,low_lim,high_lim);
        
        start_points = good_fit_pts;

        %Plot points in 5d parameter space. 
        figure(1)
        clf
        set(gcf, 'Position',[400, 400,800,300]);
        subplot(1,3,1);
        scatter3(good_fit_pts(:,1),good_fit_pts(:,2),good_fit_pts(:,3),10,...
            final_scores(good_fit_filter),'filled');hold on;
        axis([low_lim(1) high_lim(1) low_lim(2) high_lim(2) low_lim(3) high_lim(3)]);
        view(-90,0)
    
        subplot(1,3,2);
        ki_obs_log = log10(10.^good_fit_pts(:,1).*10.^good_fit_pts(:,2)./(10.^good_fit_pts(:,2)+10.^good_fit_pts(:,3)));
        scatter(ki_obs_log,...
            good_fit_pts(:,4),10,final_scores(good_fit_filter),'filled');hold on;
        
        ylim([low_lim(4) high_lim(4)]);

        xlabel('kiobs')
        ylabel('kd')
        grid on 
        grid minor


        subplot(2,3,3)

        scatter(good_fit_pts(:,5),...
            good_fit_pts(:,4),10,final_scores(good_fit_filter),'filled');hold on;
        ylim([low_lim(4) high_lim(4)]);
        xlim([low_lim(5) high_lim(5)]);
        xlabel('v');ylabel('kd');
        

        subplot(2,3,6);
        histogram(final_scores,20,'Normalization','Probability');hold on;
        xlabel('Objective function scores');
        ylabel('Count')
        refresh
        
        %SET BREAKPOINT HERE TO PAUSE BETWEEN STAGES
        disp('');
    end
    
    output_cell{i} = final_pop;
    fval_cell{i} = final_scores;
    
end


save(sprintf('search_results/%s.mat',filename),...
    'output_point','fval_point','output_cell','fval_cell',...
    'low_lim','high_lim',...
    'p_phys_stoch','p_sim_stoch',...
    'num_generations','Nuc','Cells',...
    'run_time',...
    'radius','multiplier',...
    'search_step_params','search_step_fvals','search_step_target','search_step_mode',...
    'opt_time_bounds','opt_metric','opt_target',...
    'NUM_PROC');
return

function [err_comp,bounds_met,good_fit_pts] = err_compensator(test_pts,good_fit_pts,...
    radius,multiplier,low_lim,high_lim)

m = multiplier;
r = radius;

num_conds = size(test_pts,1);
err_comp = NaN(num_conds,1);
bounds_met = false(num_conds,1);

param_range = high_lim-low_lim;
for i = 1:num_conds
    cond = test_pts(i,:);
    
    dist_from_all = (min(abs((good_fit_pts-cond)./param_range),[],2));
    min_dist = min(dist_from_all);
    if min_dist <= r
        err_comp(i) = 0;
        bounds_met(i) = true;
    else
        %step function of exponential well for cost
        err_comp(i) = 10;
        %or: err_comp(i) = exp(m*(min_dist-r))
    end
end
return

function err = fun_search_variable(...
    kvar_log,timevec,p_phys_stoch,p_sim_stoch,...
    units,onoff,sim_time_bounds,...
    target_hist,target_mean,target_bins,...
    likeness_func,...
    NUC,CELLS,NUM_PROC,...
    weight_func,err_targ,likeness_func_type,...
    low_lim,high_lim,varargin)
global search_step_params
global search_step_fvals
global search_step_target
global search_step_mode

num_conds = size(kvar_log,1);

kvar = 10.^kvar_log;


%%%%%%%%%%%%%%
%defining ss
time_filter = timevec >= sim_time_bounds(1) & timevec <= sim_time_bounds(2);
timevec_sim = timevec(time_filter);
target_hist_filter = squeeze(target_hist(time_filter,:,1:3));
target_mean_filter = target_mean(time_filter,1:3);
num_sim_t_pts = length(timevec_sim);

p_sim_stoch.CellsPerSet = CELLS;
p_sim_stoch.Nuc = NUC;

%%%%%%%%%%%%%%%%%%%%%%%%%
%CALL TO PARALLELIZER%

[nm,cm,tm,test_mesh] = parallelizer(...
    kvar,...
    timevec_sim,...
    num_conds,CELLS,...
    p_phys_stoch,p_sim_stoch,...
    onoff,units,...
    NUM_PROC);



%%%%%%
%%%%%%%
% Binning
hist_bins = target_bins;

test_histarray = NaN(num_conds,num_sim_t_pts,length(hist_bins)-1,3); 
test_meanarray = NaN(num_conds,num_sim_t_pts,3);

for cond = 1:num_conds
    nas_spec = nm(((cond-1)*CELLS+1):(cond*CELLS),:); 
    cyt_spec = cm(((cond-1)*CELLS+1):(cond*CELLS),:); 
    tot_spec = tm(((cond-1)*CELLS+1):(cond*CELLS),:); 
    for t_pt = 1:num_sim_t_pts
        test_histarray(cond,t_pt,:,1) = histcounts(nas_spec(:,t_pt),...
            hist_bins,'Normalization','probability');
        test_histarray(cond,t_pt,:,2) = histcounts(cyt_spec(:,t_pt),...
            hist_bins,'Normalization','probability');
        test_histarray(cond,t_pt,:,3) = histcounts(tot_spec(:,t_pt),...
            hist_bins,'Normalization','probability');

        test_meanarray(cond,t_pt,1) = mean(nas_spec(:,t_pt));
        test_meanarray(cond,t_pt,2) = mean(cyt_spec(:,t_pt));
        test_meanarray(cond,t_pt,3) = mean(tot_spec(:,t_pt));
    end
end

Tend = timevec(end);


    
if strcmp(likeness_func_type,'cdf')
    err = likeness_func(...
        test_histarray,target_hist_filter,...
        num_sim_t_pts,weight_func(time_filter));
elseif strcmp(likeness_func_type,'mean')
    err = likeness_func(...
        test_meanarray,target_mean_filter);
end

err = err(:,err_targ);

if ~isempty(varargin)
    discount_func = varargin{1};
    err_comp = discount_func(kvar_log);
    err = err + err_comp;
end

search_step_params{end+1} = kvar_log;
search_step_fvals{end+1} = err;
search_step_target(end+1) = err_targ;
search_step_mode{end+1} = likeness_func_type;

figure(1)
clf
set(gcf, 'Position',[400, 400,800,300]);

subplot(1,3,1)
scatter3(kvar_log(:,1),kvar_log(:,2),kvar_log(:,3),10,err,'filled');hold on;
axis([low_lim(1) high_lim(1) low_lim(2) high_lim(2) low_lim(3) high_lim(3)]);
xlabel('kinit')
ylabel('kon')
zlabel('koff')
view(-90,0)
grid on 
grid minor

subplot(1,3,2)
ki_obs_log = log10(10.^kvar_log(:,1).*10.^kvar_log(:,2)./(10.^kvar_log(:,2)+10.^kvar_log(:,3)));
if ~all(isreal(ki_obs_log(:)))
    error('?????')
    
end
scatter(ki_obs_log,...
    kvar_log(:,4),10,err,'filled');hold on;
ylim([low_lim(4) high_lim(4)]);
xlabel('kiobs')
ylabel('kd')
grid on 
grid minor

subplot(2,3,3)
scatter(kvar_log(:,5),...
    kvar_log(:,4),10,err,'filled');hold on;
xlim([low_lim(5) high_lim(5)]);
ylim([low_lim(4) high_lim(4)]);

xlabel('v');
ylabel('kd')

subplot(2,3,6)

n_col = 20;
colmap = parula(n_col);

colmap_err = linspace(min(err),max(err),n_col);
for i = 1:num_conds

    try
        col = [interp1(colmap_err,colmap(:,1),err(i)) ...
            interp1(colmap_err,colmap(:,2),err(i)) ...
            interp1(colmap_err,colmap(:,3),err(i))];
    catch
        col = [0 0 0];
    end
    
    plot(timevec_sim,test_meanarray(i,:,err_targ),...
        'Color',col);hold on;
end
scatter(timevec_sim,target_mean(time_filter,err_targ),10,'r','filled');hold on;
xlabel('Time');ylabel('Observable mean')

refresh
%SET BREAKPOINT HERE TO SEE EACH STEP'S RESULTS
return

%Feed the parameters into Gillespie in a parallel fasion - each core gets
%separate cells.
function [nm,cm,tm,testm] = parallelizer(...
    kvar,...
    timevec,...
    num_conds,Ntraj_per_set,...
    p_phys_stoch,p_sim_stoch,...    
    onoff,units,...
    NUM_PROC)   


time_matrix = repmat(timevec,num_conds*Ntraj_per_set,1);
num_t_pts = size(timevec,2);
kin = NaN(num_conds*Ntraj_per_set,5);
num_probes = size(p_phys_stoch.probes,2);

onevec = ones(Ntraj_per_set,1);

for cond = 1:num_conds    
    kin_part = [kvar(cond,1)*onevec ...
        kvar(cond,2)*onevec ...
        kvar(cond,3)*onevec ...
        kvar(cond,4)*onevec ...
        kvar(cond,5)*onevec];
    kin(((cond-1)*Ntraj_per_set+1):(cond*Ntraj_per_set),:) = kin_part; 
end


perm_key = randperm(Ntraj_per_set*num_conds);
[~,reverse_key] = sort(perm_key);
time_matrix_perm = time_matrix(perm_key,:);
kin_perm = kin(perm_key,:);

num_sets = NUM_PROC;
a = floor(linspace(1,Ntraj_per_set*num_conds,num_sets+1));
b = zeros(Ntraj_per_set*num_conds,1);
for i = 1:num_sets
    b(a(i)) = 1;
end
c = cumsum(b,1);

nm = NaN(Ntraj_per_set*num_conds,num_t_pts,num_probes);
cm = NaN(Ntraj_per_set*num_conds,num_t_pts,num_probes);
tm = NaN(Ntraj_per_set*num_conds,num_t_pts,num_probes);
testm = NaN(Ntraj_per_set*num_conds,num_t_pts,2);
[ncell,ccell,tcell,testcell] = deal(cell(num_sets,1));

parfor parallel_set = 1:num_sets
    select = c==parallel_set;
    kin_spec = kin_perm(select,:);
    time_matrix_spec = time_matrix_perm(select,:);
    sim_props = p_sim_stoch;
    sim_props.Niter = sum(select);
    [nas_mesh,cyt_mesh,tot_mesh,test_mesh] = gillespie172(...
        time_matrix_spec,p_phys_stoch,sim_props,units,onoff,kin_spec);
    
    ncell{parallel_set} = nas_mesh;
    ccell{parallel_set} = cyt_mesh;
    tcell{parallel_set} = tot_mesh; 
    testcell{parallel_set} = test_mesh; 
end
for parallel_set = 1:num_sets
    select = c==parallel_set;
    nm(select,:,:) = ncell{parallel_set};
    cm(select,:,:) = ccell{parallel_set};
    tm(select,:,:) = tcell{parallel_set};
    testm(select,:,:) = testcell{parallel_set};
end

nm = nm(reverse_key,:,:);
cm = cm(reverse_key,:,:);
tm = tm(reverse_key,:,:);
testm = testm(reverse_key,:,:);
return

function err = likeness_calc_lsq(test_mean,target_mean)
num_comparisons = size(test_mean,1);
num_species = size(test_mean,4);
err = log10(reshape(sum((test_mean-reshape(target_mean,[1,size(target_mean)])).^2,2),...
    [num_comparisons,size(test_mean,3),num_species]));
return

function err = likeness_calc_emd(test_histout_startup,target_hist_startup,...
    num_t_pts,weight_func_val)
num_comparisons = size(test_histout_startup,1);

err_deagg_nas = NaN(num_comparisons,num_t_pts);
err_deagg_cyt = NaN(num_comparisons,num_t_pts);
err_deagg_tot = NaN(num_comparisons,num_t_pts);

        
for cond = 1:num_comparisons
    for t_pt = 1:num_t_pts
        test_nas_isol = squeeze(test_histout_startup(cond,t_pt,:,1))';
        test_cyt_isol = squeeze(test_histout_startup(cond,t_pt,:,2))';
        test_tot_isol = squeeze(test_histout_startup(cond,t_pt,:,3))';

        target_nas_isol = squeeze(target_hist_startup(t_pt,:,1));
        target_cyt_isol = squeeze(target_hist_startup(t_pt,:,2));
        target_tot_isol = squeeze(target_hist_startup(t_pt,:,3));
        
        err_nas_isol = sum(abs(cumsum(test_nas_isol)-cumsum(target_nas_isol)));
        err_cyt_isol = sum(abs(cumsum(test_cyt_isol)-cumsum(target_cyt_isol)));
        err_tot_isol = sum(abs(cumsum(test_tot_isol)-cumsum(target_tot_isol)));
        
        err_deagg_nas(cond,t_pt) = err_nas_isol;
        err_deagg_cyt(cond,t_pt) = err_cyt_isol;
        err_deagg_tot(cond,t_pt) = err_tot_isol;
    end
end


weight_err = weight_func_val;
err_deagg_nas = weight_err.*err_deagg_nas;
err_deagg_cyt = weight_err.*err_deagg_cyt;
err_deagg_tot = weight_err.*err_deagg_tot;

err_nas = mean(err_deagg_nas,2);
err_cyt = mean(err_deagg_cyt,2);
err_tot = mean(err_deagg_tot,2);

err = log10([err_nas err_cyt err_tot]);
return

