function gillespie_target_generator_ex1
clear
clc
close all

delay = 0;
onoff = 'on';

p_sim_stoch = struct; 
p_sim_stoch.Nuc = ceil(5301/15);
p_sim_stoch.CellsPerSet = 10000;
% p_sim_stoch.Teq = 500; %seconds
probes = [0;0.2415];
num_probes = size(probes,2);

k_example = load('search_target_data/k_modul_ex1.mat');
kin = 10.^k_example.K_EXAMPLE(5,:);

num_t_pts = 18;
timevec = linspace(0,15,num_t_pts);

p_sim_stoch.elongationmode = 'constant';
max_bin = 100;
p_sim_stoch.max_bin = max_bin;
units = 'min';

p_phys_stoch = struct;
p_phys_stoch.L = 5300;  %nt

p_phys_stoch.probes = probes;
p_phys_stoch.ngenes = 1;
p_phys_stoch.delay = delay; %whatever the units


tic
[nas_mesh,tot_mesh,cyt_mesh,test_mesh] = parallelizer(timevec,p_phys_stoch,p_sim_stoch,units,onoff,kin);
toc

binout = (-1:max_bin)+0.5;

histout = NaN(num_t_pts,max_bin+1,3); %nascent cytosolic total
meanout = NaN(num_t_pts,3);

histout_cell = NaN(num_t_pts,max_bin+1,3);
meanout_cell = NaN(num_t_pts,3);
probe=1;
for t_pt = 1:num_t_pts
%     for probe = 1:num_probes
        histout(t_pt,:,1) = histcounts(nas_mesh(:,t_pt,probe,1),binout,'Normalization','probability');
        histout(t_pt,:,2) = histcounts(cyt_mesh(:,t_pt,probe),binout,'Normalization','probability');
        histout(t_pt,:,3) = histcounts(tot_mesh(:,t_pt,probe),binout,'Normalization','probability');
        
        meanout(t_pt,1) = mean(nas_mesh(:,t_pt,probe,1));
        meanout(t_pt,2) = mean(cyt_mesh(:,t_pt,probe));
        meanout(t_pt,3) = mean(tot_mesh(:,t_pt,probe));
%     end
    
    histout_cell(t_pt,:,1) = histcounts(test_mesh(:,t_pt,1),binout,'Normalization','probability');
    histout_cell(t_pt,:,2) = histcounts(test_mesh(:,t_pt,2),binout,'Normalization','probability');
    histout_cell(t_pt,:,3) = histcounts(test_mesh(:,t_pt,1)+test_mesh(:,t_pt,2),...
        binout,'Normalization','probability');

    meanout_cell(t_pt,1) = mean(test_mesh(:,t_pt,1));
    meanout_cell(t_pt,2) = mean(test_mesh(:,t_pt,2));
    meanout_cell(t_pt,3) = meanout_cell(t_pt,1)+meanout_cell(t_pt,2);
end

save('search_target_data/gill_target_lq.mat','binout','histout','meanout','kin',...
    'units','p_phys_stoch','p_sim_stoch','timevec','nas_mesh','tot_mesh',...
    'meanout_cell','histout_cell','test_mesh');
return


function [nm,tm,cm,testm] = parallelizer(timevec,p_phys_stoch,p_sim_stoch,...
    units,onoff,kvar)
num_conds = size(kvar,1);
Ntraj_per_set = p_sim_stoch.CellsPerSet;

time_matrix = repmat(timevec,num_conds*Ntraj_per_set,1);
num_t_pts = size(timevec,2);
num_probes = size(p_phys_stoch.probes,2);
kin = repmat(kvar,Ntraj_per_set,1);

perm_key = randperm(Ntraj_per_set*num_conds);
[~,reverse_key] = sort(perm_key);
time_matrix_perm = time_matrix(perm_key,:);
kin_perm = kin(perm_key,:);

num_sets = 72;
a = floor(linspace(1,Ntraj_per_set*num_conds,num_sets+1));
b = zeros(Ntraj_per_set*num_conds,1);
for i = 1:num_sets
    b(a(i)) = 1;
end
c = cumsum(b,1);

nm = NaN(Ntraj_per_set*num_conds,num_t_pts,num_probes);
tm = NaN(Ntraj_per_set*num_conds,num_t_pts,num_probes);
cm = NaN(Ntraj_per_set*num_conds,num_t_pts,num_probes);
testm = NaN(Ntraj_per_set*num_conds,num_t_pts,2);
[ncell,tcell,ccell,testcell] = deal(cell(num_sets,1));
parfor parallel_set = 1:num_sets
    select = c==parallel_set;
    kin_spec = kin_perm(select,:);
    time_matrix_spec = time_matrix_perm(select,:);
    sim_props = p_sim_stoch;
    sim_props.Niter = sum(select);
     [nas_mesh,cyt_mesh,tot_mesh,test_mesh] = gillespie172(...
        time_matrix_spec,p_phys_stoch,sim_props,units,onoff,kin_spec);
    
    ncell{parallel_set} = nas_mesh;
    tcell{parallel_set} = tot_mesh; 
    ccell{parallel_set} = cyt_mesh; 
    testcell{parallel_set} = test_mesh; 
end

for parallel_set = 1:num_sets
    select = c==parallel_set;
    nm(select,:,:) = ncell{parallel_set};
    tm(select,:,:) = tcell{parallel_set};
    cm(select,:,:) = ccell{parallel_set};
    testm(select,:,:) = testcell{parallel_set};
end

nm = nm(reverse_key,:,:);
tm = tm(reverse_key,:,:);
cm = cm(reverse_key,:,:);
testm = testm(reverse_key,:,:);
return